#!/bin/bash 
# ================================================== #
#               Load Balancer Helper                 #
#                                                    #
#  Linux "ipvs/ipvsadm" implementation               #
#                                                    #
  VERSION="1.3"                                      #
# ================================================== #
#
# Parameters (optional):
# -i IPINTF -p IPVSPERSIST -j IPVSSYNCINTF -s IPVSSYNCID -t READTIMEOUT -f VIPADDRFILE -r RATIO -m
# (-m means 'model': no real OS command is executed)
#

IPINTF="eth0"               # the Interface on which we need to arp-respond to VIPs
READTIMEOUT=15              # read time-out: if expired, we are sending "* ping"
IPVSPERSIST=15              # the IPVS persistancy time period
IPVSSYNCID="0"              # the IPVS daemon syncing parameter
LOCALRATIO="75"             # the relative (to 100) weight of the local server
EXECCMD="YES"               # execute the shell commands
VIPADDRFILE="/etc/sysconfig/vipaddrs"
                            # file with VIP (Virtual IP) addresses
IPFMARK="1000"              # the iptables marker value for the incoming VIP packets

IPVSVIRT="-s rr"            # the IPVS parameter for the virtual service
IPVSSYNCINTF=""             # the Interface for daemon syncing packets: empty -> IPINTF

#
# internal variables
#
isActive=0                  # set to 1 when the active balancer is activated
doOutput=1                  # set to 0 on a signal, so we do not try to write to stdout again
localMember="0.0.0.0"       # the local IP of this server
                            # when not set to 0.0.0.0, we start the balancer and make it route all traffic to it
IPMANGLESNEXT=0             # the first unused index of the iptables -t mangle rules

declare -A currentMembers   # current   balancer config: member.ip -> params list
declare -A reportedMembers  # requested balancer config: member.ip -> params list
declare -A mangleRules      # member.ip -> index of the "tcp response packets" marking rule in iptables -t mangle 

while getopts "i:f:p:j:s:t:r:m" OPTION; do
  case $OPTION in
  i) IPINTF=$OPTARG ;;
  f) VIPADDRFILE=$OPTARG ;;
  p) IPVSPERSIST=$OPTARG ;;
  j) IPVSSYNCINTF=$OPTARG ;;
  s) IPVSSYNCID=$OPTARG ;;
  t) READTIMEOUT=$OPTARG ;;
  r) LOCALRATIO=$OPTARG ;;
  m) EXECCMD="NO" ;;
  esac
done

if [ "$IPVSPERSIST" != "0" ]; then IPVSVIRT="$IPVSVIRT -p $IPVSPERSIST"; fi
if [ "$IPVSSYNCINTF" == "" ]; then IPVSSYNCINTF="$IPINTF"; fi

PATH=/sbin:$PATH
VIPLIST=""                     # list of VIP masks: xxx.yy.zz.tt/mm where mm should be >= 24

readVIPList() {
  local xVIP
  for xVIP in `cat $VIPADDRFILE | grep -v '^#'`; do
    if [[ $xVIP != */* ]]; then xVIP=$xVIP/32; fi
    if (( ${xVIP##*/} < 24)); then
      echo "* ERROR: incorrect mask: $xVIP"; exit 1;
    fi
    VIPLIST="$VIPLIST$xVIP "
  done
}

#
# set lastIPByteValue to "yyy" from "xxx.xxx.xxx.yyy"
#
getLastIPByte() {
  lastIPByteValue=$1
  lastIPByteValue=${lastIPByteValue##*.}
}

#
# shell command evaluation:
#   output the command to the CommuniGate Pro log (using the "* command" response to stdout)
#   execute the command, storing its stdout and stderr into a variable
#   output the command result/error (if any) to the CommuniGate Pro Log
#
execCommand() {
  local resultOut
  if [ $doOutput != 0 ]; then echo "* $1"; fi
  resultOut=""
  if [ $EXECCMD == "YES" ]; then
    resultOut=`$1 2>&1 | tr '\n' ' '`
  fi
  if [ "x$resultOut" != "x" ] && [ $doOutput != 0 ]; then echo "* result=$resultOut"; fi
}

#
# Removing a load balancer group member ($1 is its "WAN address")
#
removeMember() {
  local xMember="$1"
  execCommand "ipvsadm -d -f $IPFMARK -r $xMember"              # remove this member from the load balancer set

  getLastIPByte $xMember
  local ipfResponseMark=$((IPFMARK+lastIPByteValue))
  execCommand "ipvsadm -D -f $ipfResponseMark"                  # remove the IPVS routing for response packets

  local value=${mangleRules[$xMember]}
  execCommand "iptables -t mangle -D PREROUTING $value"         # remove the iptables rule for response packets,
  unset currentMembers[$xMember]                                # and renumber all rules that were after the removed one
  IPMANGLESNEXT=$((IPMANGLESNEXT-1))                            # @1.0

  local x
  for x in ${!mangleRules[*]}; do
    if (( ${mangleRules[$x]} > $value )); then
      mangleRules[$x]=$((mangleRules[$x]-1))
    fi
  done
}

#
# Set the Load Balancer "real servers"
# Associative array activeMembers has the current config:  {ip.addr = (params)}, where params = (YES/NO, tcpFrom, tcpTill, ratio)
# Associative array reportedMembers has the new config
#
setReportedMembers() {
  local xMember oldInfo newInfo command ipfResponseMark ratio

  # add or update reportedMembers elements
  for xMember in "${!reportedMembers[@]}"; do
    newInfo=(${reportedMembers[$xMember]})
    oldInfo=(${currentMembers[$xMember]})

    getLastIPByte $xMember
    ipfResponseMark=$((IPFMARK+lastIPByteValue))      # mark response packet for xx.xx.xx.yyy server with 1000+yyy marker
    command=""
    if [ "${#oldInfo}" == 0 ]; then
      command="-I"
      mangleRules[$xMember]=$IPMANGLESNEXT
      IPMANGLESNEXT=$((IPMANGLESNEXT+1))

      execCommand "ipvsadm -A -f $ipfResponseMark"
      execCommand "ipvsadm -a -f $ipfResponseMark -r $xMember -g"
    elif [ "${oldInfo[1]}:${oldInfo[2]}" != "${newInfo[1]}:${newInfo[2]}" ]; then
      command="-R"
    fi

    if [ "$command" != "" ]; then                     # create or update the iptables rule for the specified port range
      execCommand "iptables -t mangle $command PREROUTING ${mangleRules[$xMember]} -p tcp -m tcp --dport ${newInfo[1]}:${newInfo[2]} -m mark --mark $IPFMARK -j MARK --set-mark $ipfResponseMark" 
    fi

    if [[ $isActive == 0 && $xMember != $localMember ]]; then newInfo[0]="NO"; fi

    command=""
    if [ "${#oldInfo}" == 0 ]; then
      command="-a"
    elif [ "${newInfo[0]}" != "${oldInfo[0]}" ]; then
      command="-e"
    elif [ "X${newInfo[3]}" != "X${oldInfo[3]}" ]; then
      command="-e"
    fi

    if [[ "$command" != "" ]]; then
      command="ipvsadm $command -f $IPFMARK -r $xMember -g"
      ratio="100"
      if [ "${newInfo[0]}" != "YES" ]; then
        ratio="0"
      elif (( "${#newInfo[@]}" >= 4 )); then
        ratio="${newInfo[3]}"
      fi
      if [ $xMember == $localMember ]; then
        ratio=$(($ratio * $LOCALRATIO / 100))
      fi

      execCommand "$command -w $ratio"
    fi

    currentMembers[$xMember]="${newInfo[*]}"

  done

  # remove all currentMembers that do not exist in reportedMembers
  for xMember in "${!currentMembers[@]}"; do
    if [ "${reportedMembers[$xMember]}" == "" ]; then
      removeMember $xMember
    fi
  done
}


#
# Setting ARP response mode:
#  if the parameter is 0, we disable arp responses for our VIPs (created on the "lo" interface)
#  if the parameter is 1, we enable them
#
setARPVIPResponse() {
  if [ $1 == 0 ]; then
    execCommand "sysctl net.ipv4.conf.all.arp_ignore=1"
    execCommand "sysctl net.ipv4.conf.$IPINTF.arp_ignore=1"
    execCommand "sysctl net.ipv4.conf.all.arp_announce=2"
    execCommand "sysctl net.ipv4.conf.$IPINTF.arp_announce=2"
  else
    execCommand "sysctl net.ipv4.conf.all.arp_ignore=0"
    execCommand "sysctl net.ipv4.conf.$IPINTF.arp_ignore=0"
    execCommand "sysctl net.ipv4.conf.all.arp_announce=0"
    execCommand "sysctl net.ipv4.conf.$IPINTF.arp_announce=0"
  fi
}

#
# For each VIP address we send an ARP broadcast
# binding that address to our $IPINT interface MAC
#  we repeat this twice to ensure that the router gets it
#
broadcastARPVIP() {
  local x xVIP xIP xIP0 xIP1 xMask index
  for x in 1 2; do
    for xVIP in $VIPLIST; do
      xIP=${xVIP%/*}                          # xx.xx.xx.yy/mm -> xx.xx.xx.yy
      xIP0=${xIP%.*}                          # xx.xx.xx.yy/mm -> xx.xx.xx
      xIP1=${xIP##*.}                         # xx.xx.xx.yy/mm -> yy
      xMask=$(( 2 ** (32 - ${xVIP##*/}) ))
      index=0

      if [[ $xIP0 == N* ]]; then              # Nxx.xx.xx.yy/mm  -> remove "N", exclude the first (network), second (router), and last (broadcast) addresses
        xIP0=${xIP0##N}
        index=2; xMask=$(($xMask - 1))
      fi
      for ((; index<$xMask; index++ )); do
        execCommand "arping -c 1 -I $IPINTF -U $xIP0.$((xIP1 + index))"
      done
    done
  done
}

# Init everything
#  we start as a backup balancer
#
initiate() {
  local xVIP index
  index=1
  for xVIP in $VIPLIST; do
    xVIP=${xVIP##N}
    execCommand "iptables -t raw    -I PREROUTING $index -d $xVIP -j NOTRACK"                   # switch off connection tracking for VIPs
    execCommand "iptables -t mangle -I PREROUTING $index -d $xVIP -j MARK --set-mark $IPFMARK"  # mark all packets directed to VIPs
    index=$((index+1))
  done
  IPMANGLESNEXT=$index

  execCommand "ipvsadm -A -f $IPFMARK $IPVSVIRT"       # start the balancer now, otherwise we will have problems with the Backup Daemon filling our table
  startDaemon backup
}

#
# Cleanup everything: on exit, or we stop being an active balancer:
#  disable ARP responses for VIP (we should do this first to let the other balancer take over)
#  restore the IPVS config by removing the load balancer entry
#
cleanup() {
  if [ $isActive != 0 ]; then
    isActive=0
    setARPVIPResponse 0                                 # stop answering to VIP arp requests
    stopDaemon master
  else
    stopDaemon backup
  fi

  local xMember xVIP
  for xMember in ${!currentMembers[*]}; do
    removeMember $xMember
  done

  execCommand "ipvsadm -D -f $IPFMARK"                  # remove the balancer entry

  for xVIP in $VIPLIST; do
    xVIP=${xVIP##N}
    execCommand "iptables -t raw    -D PREROUTING 1"    # delete the inserted rules
    execCommand "iptables -t mangle -D PREROUTING 1"    # delete all marker rules
  done

}

startDaemon() {
  if [ $1 == "backup" ] && sysctl net.ipv4.vs.backup_only 2>/dev/null 1>/dev/null ; then
    execCommand "sysctl net.ipv4.vs.backup_only=1"
  fi
  execCommand "ipvsadm --start-daemon $1 --mcast-interface $IPVSSYNCINTF --syncid $IPVSSYNCID"
}

stopDaemon() {
  execCommand "ipvsadm --stop-daemon $1"
  if [ $1 == "backup" ] && sysctl net.ipv4.vs.backup_only 2>/dev/null 1>/dev/null ; then
    execCommand "sysctl net.ipv4.vs.backup_only=0"
  fi
}


#
# If we got an interruption: the CommuniGate Pro server has quitted and closed the output pipe
#
on_die() {
  doOutput=0
  cleanup
  exit 0
}

echo "* helper v$VERSION started"
trap 'on_die' SIGPIPE SIGTERM

readVIPList
initiate

#
# for pre-4.0 bash, we rely on the fact that time-out does not change the PREFIX value, while EOF sets it to and emty string
#
while true; do
  PREFIX="*"
  RESPONSE="OK"

  read -r -t $READTIMEOUT PREFIX CMD PARAM1 PARAM2
  retCode=$?
  if (( "$retCode" >= 128 )) || [ "$PREFIX" == "*" ]; then
    echo "* ping"

  elif(( "$retCode" == 0 )); then

    case $CMD in
    "INTF" )
      RESPONSE="INTF 1"
      ;;
    "QUIT" )
      echo "* helper v$VERSION quitting"
      cleanup
      echo "$PREFIX $RESPONSE"
      exit 0
      ;;
    "LOCAL" )
      reportedLocal=${PARAM1//[\"\[\]]/}
      if [ "$reportedLocal" != "$localMember" ]; then
        if [ "$localMember" != "0.0.0.0" ]; then removeMember $localMember; fi
        localMember="$reportedLocal"
        setReportedMembers
#       if [ "$localMember" != "0.0.0.0" ]; then execCommand "ipvsadm -a -f $IPFMARK -r $localMember -g -w $LOCALRATIO"; fi
      fi
      ;;
    "MEMBERS" )                                               # {address=(value1,value2,value3);address=(....);....}
      unset reportedMembers
      declare -A reportedMembers
      for x in ${PARAM1//[\{\}\;]/ }; do
        value="${x#*=}"
        reportedMembers[${x%%=*}]=${value//[()#,]/ }          # reportedMembers[address] = "value1 value2 value3"
      done
      setReportedMembers
      ;;
    "STARTBALANCER" )
      if [ $isActive == 0 ]; then
        echo "* switching on"
        isActive=1
        setReportedMembers
        stopDaemon  backup
        startDaemon master
        setARPVIPResponse 1
        broadcastARPVIP
      fi
      ;;
    "STOPBALANCER" )
      if [ $isActive != 0 ]; then
        echo "* switching off"
        setARPVIPResponse 0                             # stop answering to VIP arp requests
        isActive=0
        setReportedMembers                              # restore the IPVS software to the initial mode
        stopDaemon  master
        startDaemon backup
      fi
      ;;
    * )
      RESPONSE="ERROR \"unknown command: $CMD\""
      ;;
    esac
    echo "$PREFIX $RESPONSE"
  else
    echo "* stopping"
    cleanup
    exit 0
  fi
done
